/* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* mysqldump.c  - Dump a tables contents and format to an ASCII file
**
** The author's original notes follow :-
**
**		******************************************************
**		*						     *
**		* AUTHOR: Igor Romanenko (igor@frog.kiev.ua)	     *
**		* DATE:   December 3, 1994			     *
**		* WARRANTY: None, expressed, impressed, implied      *
**		*	    or other				     *
**		* STATUS: Public domain				     *
**		* Adapted and optimized for MySQL by		     *
**		* Michael Widenius, Sinisa Milivojevic, Jani Tolonen *
**		* -w --where added 9/10/98 by Jim Faucette	     *
**		* slave code by David Saez Padros <david@ols.es>     *
**		*						     *
**		******************************************************

/******************************************************************
**
** modification by:
** DATATEC Computer u. Elektronik GmbH
** DumpFunc.cpp for BackUp.cpp DLL version 1.0
**
** Author Ch. Friedrich 20.5.2001
**
** Dependencies: libmysql.dll
** Output:       MyBackUp.dll
**
*******************************************************************/

#include "DumpFunc.h"


int TMyDumpObj::init_dumping(FILE *sql_file,char *database)
{
  TbRecords = 0L;
  if (mysql_select_db(My_Sock, database))
  {
    return 1; // fails
  }
  if (DpO.opt_databases)
  {
    // first write Version-Info !!
    fprintf(sql_file, "#%s\n", DMPVERSION);
    fprintf(sql_file,"\n# Current Database: %s\n#\n", database);
    if (!DpO.opt_create_db)
      fprintf(sql_file,"\nCREATE DATABASE /*!32312 IF NOT EXISTS*/ %s;\n", database);
    fprintf(sql_file,"\nUSE %s;\n", database);
  }
  return 0;
} /* init_dumping */


void TMyDumpObj::write_heder(FILE *sql_file, char *db_name)
{
 if(DpO.Write_heder)
 {
  fputs("# MySQL (Win) Dump \n#\n",sql_file);
  fprintf(sql_file, "# Host: %s    Database: %s\n",
	  current_host ? current_host : "localhost", db_name ? db_name : "");
  fputs("#-------------------------------------------------------------\n",
  sql_file);
  fprintf(sql_file, "# Server version\t%s\n",
	  mysql_get_server_info(My_Sock));
 } else fprintf(sql_file, "\n");
 return;
}


void TMyDumpObj::safe_exit(int error)
{
  if(SQLRes)
   mysql_free_result(SQLRes);
}


int TMyDumpObj::unescape(FILE *file,char *pos,int length)
{
  char *tmp;
  if (!(tmp=(char*) malloc(length*2+1)))
  {
    return 0;				/* Fatal error */
  }
  mysql_escape_string(tmp, pos, length);
  fputc('\'', file);
  fputs(tmp, file);
  fputc('\'', file);
  free(tmp);
  return 1;				/* Fatal error */
} /* unescape */


int TMyDumpObj::test_if_special_chars(const char *str)
{
#if MYSQL_VERSION_ID >= 32300
  for ( ; *str ; str++)
    if (!isvar(*str) && *str != '$')
      return 1;
#endif
  return 0;
} /* test_if_special_chars */

char * TMyDumpObj::strmov(char *dst, const char *src)
{
  while ((*dst++ = *src++)) ;
  return dst-1;
}

char * TMyDumpObj::strend( const char *s)
{
  while (*s++);
  return (char*) (s-1);
}

char * TMyDumpObj::quote_name(char *name, char *buff)
{
  char *end;
  if (!DpO.opt_quoted && !test_if_special_chars(name))
    return name;
  buff[0]=QUOTE_CHAR;
  end=strmov(buff+1,name);
  end[0]=QUOTE_CHAR;
  end[1]=0;
  return buff;
} /* quote_name */


/*
** getStructure -- retrievs database structure, prints out corresponding
**       CREATE statement and fills out insert_pat.
** Return values:  number of fields in table, 0 if error
*/
int TMyDumpObj::getTableStructure(MYSQL *sock,char *table, char* db, FILE  *sql_file)
{
  PMYSQL_ROW  row;
  int init=0;
  uint       numFields;
  char 	     *strpos;
  const char *delayed;
  char 	     name_buff[NAME_LEN+3],table_buff[NAME_LEN+3];

  delayed= "";
  My_Sock=sock;

  sprintf(insert_pat,"show fields from %s",table);
  if (mysql_query(sock,insert_pat) || !(SQLRes=mysql_store_result(sock)))
  {
      safe_exit(EX_MYSQLERR);
      return 0;
    }

    // first write Version-Info !!
    fprintf(sql_file, "#%s\n", DMPVERSION);
    /* Make an sql-file, if path was given iow. option -T was given */
    if (!DpO.tFlag)
    {
      write_heder(sql_file, db);
      fprintf(sql_file, "\n# Table structure for table '%s'\n#\n", table);
      if (DpO.opt_drop)
        fprintf(sql_file, "DROP TABLE IF EXISTS %s;\n",table);
      fprintf(sql_file, "CREATE TABLE %s (\n", table);
    }


    if (DpO.cFlag)
      sprintf(insert_pat, "INSERT %sINTO %s (", delayed, table);
    else
    {
      sprintf(insert_pat, "INSERT %sINTO %s VALUES ", delayed, table);
      if (!DpO.extended_insert) strcat(insert_pat,"(");
    }

    strpos=strend(insert_pat);
    while ((row=mysql_fetch_row(SQLRes)))
    {
      int *lengths=mysql_fetch_lengths(SQLRes);
      if (init)
      {
        if (!DpO.tFlag)
		fputs(",\n",sql_file);
        if (DpO.cFlag)
		strpos=strmov(strpos,", ");
      }
      init=1;
      if (DpO.cFlag)
        strpos=strmov(strpos,quote_name(row[SHOW_FIELDNAME],name_buff));
      if (!DpO.tFlag)
      {
        if (DpO.opt_keywords)
		fprintf(sql_file, "  %s.%s %s", table,
			quote_name(row[SHOW_FIELDNAME],name_buff), row[SHOW_TYPE]);
        else
		fprintf(sql_file, "  %s %s", quote_name(row[SHOW_FIELDNAME],name_buff),
			row[SHOW_TYPE]);
        if (row[SHOW_DEFAULT])
        {
		fputs(" DEFAULT ", sql_file);
		unescape(sql_file,row[SHOW_DEFAULT],lengths[SHOW_DEFAULT]);
        }
        if (!row[SHOW_NULL][0])
		fputs(" NOT NULL", sql_file);
        if (row[SHOW_EXTRA][0])
		fprintf(sql_file, " %s",row[SHOW_EXTRA]);
      }
    }
    numFields = mysql_num_rows(SQLRes);
    mysql_free_result(SQLRes);
    if (!DpO.tFlag)
    {
      /* Make an sql-file, if path was given iow. option -T was given */
      char buff[20+FN_REFLEN];
      int keynr,primary_key;
      sprintf(buff,"show keys from %s",table);
      if (mysql_query(My_Sock, buff))
      {
        safe_exit(EX_MYSQLERR);
        return 0;
      }
      SQLRes=mysql_store_result(My_Sock);
      /* Find first which key is primary key */
      keynr=0;
      primary_key=INT_MAX;
      while ((row=mysql_fetch_row(SQLRes)))
      {
        if (atoi(row[3]) == 1)
        {
          keynr++;
          if (!strcmp(row[2],"PRIMARY"))
          {
            primary_key=keynr;
            break;
          }
        }
      }
      mysql_data_seek(SQLRes,(my_ulonglong)0);
      keynr=0;
      while ((row=mysql_fetch_row(SQLRes)))
      {
        if (atoi(row[3]) == 1)
        {
		if (keynr++)
                  fputs(")",sql_file);
		if (atoi(row[1]))       /* Test if duplicate key */
		  /* Duplicate allowed */
		  fprintf(sql_file, ",\n  KEY %s (",quote_name(row[2],name_buff));
		else if (keynr == primary_key)
		  fputs(",\n  PRIMARY KEY (",sql_file); /* First UNIQUE is primary */
		else
		  fprintf(sql_file, ",\n  UNIQUE %s (",quote_name(row[2],name_buff));
        }
        else fputs(",",sql_file);

        fputs(quote_name(row[4],name_buff), sql_file);
        if (row[7]) fprintf(sql_file, " (%s)",row[7]);      /* Sub key */
      }
      if (keynr) fputs(")",sql_file);
      /* Get MySQL specific create options - ausgel.*/
      fputs("\n)",sql_file);
      fputs(";\n", sql_file);
  }
  if (DpO.cFlag)
  {
    strpos=strmov(strpos,") VALUES ");
    if (!DpO.extended_insert) strpos=strmov(strpos,"(");
  }
  return numFields;
} /* getTableStructure */



void  TMyDumpObj::dumpTable(int numFields, char *table, FILE  *sql_file)
{
  char query[1024], *end, buff[256],table_buff[NAME_LEN+3];
  MYSQL_FIELD  *field;
  PMYSQL_ROW    row;
  ulong		NumRows, init_length, line_length,line_count;
  ulong net_buffer_length=8192;	/* Default length. Enlarged if necessary */
  char ProgrAktiv=0;
  uint Step;

    if (MyProgress) ProgrAktiv=1;
    TbRecords = 0L;

    if (ProgrAktiv)
    {
      MyProgress->StepIt() ;
      Application->ProcessMessages();
    }
    // first write Version-Info !!
    fprintf(sql_file, "#%s\n", DMPVERSION);
    fprintf(sql_file,"\n# Dumping data for table '%s'\n", table);
    sprintf(query, "SELECT * FROM %s", quote_name(table,table_buff));
    if (DpO.opt_where)
    {
      fprintf(sql_file,"# WHERE:  %s\n",DpO.WhereStr);
      strcat(query, " WHERE ");
      strcat(query,DpO.WhereStr);
    }
    fputs("#\n",sql_file);

    if (mysql_query(My_Sock, query))
    {
      return;
    }
    if (ProgrAktiv)
    {
      MyProgress->StepIt() ;
      Application->ProcessMessages();
    }

    SQLRes=mysql_store_result(My_Sock);
    if (!SQLRes)
    {
      safe_exit(EX_CONSCHECK);
      return;
    }
    if (mysql_num_fields(SQLRes) != numFields)
    {
      safe_exit(EX_CONSCHECK);
      return;
    }

    if (ProgrAktiv)
    {
      MyProgress->StepIt() ;
      Application->ProcessMessages();
    }

    NumRows=mysql_num_rows(SQLRes);
    if(!NumRows )
    {
      fprintf(sql_file,"\n# Sorry - the table '%s' was empty!\n", table);
      safe_exit(EX_CONSCHECK);
      return;
    }

    if (ProgrAktiv)
    {
      MyProgress->StepIt() ;
      Application->ProcessMessages();
    }

    if (DpO.opt_lock)
      fprintf(sql_file,"LOCK TABLES %s WRITE;\n", quote_name(table,table_buff));

    line_count=0;
    line_length=0;

    if (DpO.extended_insert)
    {
      line_length+=strlen(insert_pat);
      fputs(insert_pat,sql_file);
    }

    while ((row=mysql_fetch_row(SQLRes)))
    {
      int i;
      int *lengths=mysql_fetch_lengths(SQLRes);
      line_count++;
      TbRecords++;
      Step++;

      if ((Step > ProgrStep) && ( ProgrAktiv))
      {
        MyProgress->StepIt() ;
        Application->ProcessMessages();
        Step=0;
      }
      if (!DpO.extended_insert) fputs(insert_pat,sql_file);

      mysql_field_seek(SQLRes,0);

      for (i = 0; i < mysql_num_fields(SQLRes); i++)
      {
	if (!(field = mysql_fetch_field(SQLRes)))
	{
	  safe_exit(EX_CONSCHECK);
	  return;
	}
	if (DpO.extended_insert)
	{
          if ((line_count>1) && (i==0)) fputc(',',sql_file); // after first Line

          if (i==0) fputc('(',sql_file);
	  else fputc(',',sql_file);

	  if (row[i])
	  {
              line_length+=lengths[i]+2;
              if (!IS_NUM_FIELD(field)) unescape(sql_file, row[i], lengths[i]);
              else fputs(row[i],sql_file);
          }
          else
          {
            fputs("NULL",sql_file);
            line_length+=4+2;
          }
	}
	else
	{
          if (i) fputc(',',sql_file);
          if (row[i])
          {
            if (!IS_NUM_FIELD(field))
              unescape(sql_file, row[i], lengths[i]);
            else
              fputs(row[i],sql_file);
          }
          else
          {
            fputs("NULL",sql_file);
          }
        }
      }

      if (DpO.extended_insert)
      {
        fputs(")",sql_file);
        line_length++;
        if (line_length > net_buffer_length)
        {
          fputs(";\n",sql_file);
          line_count=0;
          if (TbRecords < NumRows)
          {
            line_length=strlen(insert_pat);
            fputs(insert_pat,sql_file);
          }
        }
      }
      else
      {
       fputs(");\n",sql_file);
      }
    }
    if ((DpO.extended_insert) && (line_length <= net_buffer_length))  fputs(";\n",sql_file);
    fflush(sql_file);
    if (mysql_errno(My_Sock))
    {
      safe_exit(EX_CONSCHECK);
      return;
    }
    fprintf(sql_file, "\n");
    if (DpO.opt_lock) fprintf(sql_file, "UNLOCK TABLES;\n");
    fprintf(sql_file, "\n");
    fflush(sql_file);
    mysql_free_result(SQLRes);
} /* dumpTable */

